# coding=utf-8
"""
Pytest config file
"""
import os
import sys
from pathlib import Path

import pytest

# from esst import core


def pytest_configure(config):
    """
    Runs at tests startup

    Args:
        config: pytest config args
    """
    print('pytest args: ', config.args)
    os.environ['DCS_PATH'] = 'test'
    os.environ['DCS_SERVER_NAME'] = 'test'
    os.environ['DCS_SERVER_PASSWORD'] = 'test'
    os.environ['DISCORD_BOT_NAME'] = 'test'
    os.environ['DISCORD_CHANNEL'] = 'test'
    os.environ['DISCORD_TOKEN'] = 'test'
    sys._called_from_test = True


# noinspection SpellCheckingInspection
def pytest_unconfigure(config):
    """Tear down"""
    print('pytest args: ', config.args)
    # noinspection PyUnresolvedReferences,PyProtectedMember
    del sys._called_from_test


# @pytest.fixture(autouse=True)
# def _reset_fs():
#     core.FS._reset()
#     yield
#     core.FS._reset()


@pytest.fixture(autouse=True)
def _dummy_config():
    Path('./esst.yml').write_text("""
dcs_path: './DCS'
dcs_server_name: 'server_name'
dcs_server_password: 'server_pwd'
discord_bot_name: 'bot_name'
discord_channel: 'channel'
discord_token: 'token'
    """)
    yield


@pytest.fixture(autouse=True)
def cleandir(request, tmpdir):
    """
    Creates a clean directory and cd into it for the duration of the test

    Args:
        request: Pytest request object
        tmpdir: Pytest tmpdir fixture

    """
    # from esst.core import FS
    # FS.saved_games_path = Path(str(tmpdir), 'Saved Games').absolute()
    # FS.ur_install_path = Path(str(tmpdir), 'UniversRadio').absolute()
    if 'nocleandir' in request.keywords:
        yield
    else:
        current_dir = os.getcwd()
        os.chdir(str(tmpdir))
        yield os.getcwd()
        os.chdir(current_dir)


def pytest_addoption(parser):
    """Adds options to Pytest command line"""
    parser.addoption("--long", action="store_true",
                     help="run long tests")


def pytest_runtest_setup(item):
    """Test suite setup"""

    # Skip tests that are marked with the "long" marker
    long_marker = item.get_marker("long")
    if long_marker is not None and not item.config.getoption('long'):
        pytest.skip('skipping long tests')



# coding=utf-8
"""
Manages Discord commands
"""

from esst.core import CTX, MAIN_LOGGER

LOGGER = MAIN_LOGGER.getChild(__name__)


class DISCORD:
    """
    Manages Discord commands
    """

    @staticmethod
    def say(message):
        """
        Sends a message on the active channel

        Args:
            message: string to send

        """
        CTX.discord_msg_queue.put(message[:1].upper() + message[1:])

    @staticmethod
    def send_file(file_path: str):
        """
        Sends a file to the active channel

        Args:
            file_path: path to the file to send

        """
        LOGGER.debug(f'sending file to Discord: {file_path}')
        CTX.discord_file_queue.put(file_path)

    @staticmethod
    def can_start():
        """DCS can start"""
        if not CTX.discord_can_start:
            LOGGER.debug('Discord can start')
        CTX.discord_can_start = True

    @staticmethod
    def cannot_start():
        """DCS cannot start"""
        if CTX.discord_can_start:
            LOGGER.debug('Discord can NOT start')
        CTX.discord_can_start = False



# coding=utf-8
"""
Tests for esst.core.FS
"""
from pathlib import Path

import pytest

from esst.core import FS


@pytest.fixture(autouse=True)
def _setup():
    FS.saved_games_path = './Saved Games'
    FS.dcs_path = './DCS'
    Path('./Saved Games').mkdir()
    Path('./Saved Games/DCS').mkdir()
    Path('./DCS').mkdir()


def test_ensure_path():
    FS.saved_games_path = None

    with pytest.raises(RuntimeError):
        # noinspection PyTypeChecker
        FS.ensure_path(FS.saved_games_path, 'test')

    with pytest.raises(FileNotFoundError):
        FS.ensure_path('./test', 'test')


def test_saved_games_not_found():
    Path('./Saved Games/DCS').rmdir()
    Path('./Saved Games').rmdir()
    with pytest.raises(FileNotFoundError) as exc:
        FS.get_saved_games_variant()
    assert str(exc).endswith('Saved Games')


def test_no_dcs_dir_in_saved_games():
    Path('./Saved Games/DCS').rmdir()
    with pytest.raises(FileNotFoundError) as exc:
        FS.get_saved_games_variant()
    assert str(exc).endswith('Saved Games\\DCS')


def test_no_dcs_dir():
    Path('./DCS').rmdir()
    with pytest.raises(FileNotFoundError) as exc:
        FS.get_saved_games_variant('./DCS')
    assert str(exc).endswith('\\DCS')


def test_base_variant():
    variant = FS.get_saved_games_variant()
    assert isinstance(variant, Path)
    expected = Path('Saved Games/DCS').absolute()
    assert variant.samefile(expected)


def test_variant_open_beta_missing():
    variant = Path('./DCS/dcs_variant.txt')
    variant.write_text('openbeta')

    with pytest.raises(FileNotFoundError) as exc:
        FS.get_saved_games_variant()
    assert str(exc).endswith('Saved Games\\DCS.openbeta')


def test_variant_open_beta():
    variant = Path('./DCS/dcs_variant.txt')
    variant.write_text('openbeta')
    Path('./Saved Games/DCS.openbeta').mkdir()
    assert isinstance(FS.get_saved_games_variant('./DCS'), Path)



# coding=utf-8
"""
Manages config params for Univers Radio
"""

from elib.config import ConfigProp

NAMESPACE = 'UR'


# pylint: disable=too-few-public-methods
class URConfig:
    """
    Manages config params for Univers Radio
    """

    ur_path = ConfigProp(str, default='', namespace=NAMESPACE)



# coding=utf-8
"""
Manages UR voice service executable
"""
import os
import time
from pathlib import Path

import psutil

from esst.core import FS, MAIN_LOGGER

LOGGER = MAIN_LOGGER.getChild(__name__)

PROC_NAME = 'UR_VoiceServiceServer.exe'


class URVoiceService:
    """
    Manages UR voice service executable
    """
    pid = None

    @staticmethod
    def is_running() -> bool:
        """Returns UR voice service status"""
        for proc in psutil.process_iter():
            if proc.name() == PROC_NAME:
                URVoiceService.pid = proc.pid
                return True
        return False

    @staticmethod
    def start_service():
        """
        Starts UR voice service
        """
        exe_path = Path(FS.ur_install_path, PROC_NAME)
        if not exe_path.exists():
            raise FileNotFoundError(exe_path)
        LOGGER.info(f'starting UR voice service: {exe_path}')
        os.startfile(str(exe_path))
        URVoiceService.is_running()

    @staticmethod
    def poll():
        """
        Checks that UR voice service is running
        """
        LOGGER.debug('polling UR voice service')
        proc = psutil.Process(URVoiceService.pid)
        if not proc.status() == psutil.STATUS_RUNNING:
            raise RuntimeError('UR voice service stopped')

    @staticmethod
    def kill():
        """
        Kills UR voice service
        """
        if URVoiceService.is_running():
            try:
                proc = psutil.Process(URVoiceService.pid)
                LOGGER.info('killing UR voice service')
                proc.terminate()
                while URVoiceService.is_running():
                    LOGGER.debug('waiting on UR voice service to close')
                    time.sleep(1)
            except FileNotFoundError:
                pass
            except psutil.NoSuchProcess:
                pass



# coding=utf-8
"""
Manages a mission change context
"""
from pathlib import Path

from . import store


class MissionManagerContext:
    """
    Manages a mission change context
    """

    def __init__(
            self,
            original_mission_path: Path,
            originator: str,
    ):
        self._original_path = original_mission_path
        self._originator = originator
        self._tmp_path = store.get_random_auto_mission_name(self._original_path)

    @property
    def original_path(self) -> Path:
        """
        Returns: source mission for this context
        """
        return self.original_path

    @property
    def originator(self) -> str:
        """
        Returns: name of the person who issued the command
        """
        return self._originator

    @property
    def tmp_path(self) -> Path:
        """
        Returns: temporary path for this mission
        """
        return self._tmp_path



# coding=utf-8
"""
Server machine commands
"""


import os

from esst.commands import DCS, DISCORD
from esst.core import CTX, MAIN_LOGGER
from esst.utils.historygraph import make_history_graph

LOGGER = MAIN_LOGGER.getChild(__name__)


class SERVER:
    """
    Server machine commands
    """

    @staticmethod
    def reboot(force: bool = False):
        """Reboots the server computer"""
        if DCS.there_are_connected_players():
            if not force:
                return 'there are connected players; cannot restart the server now (use "--force" to restart anyway)'
            else:
                LOGGER.warning('forcing restart with connected players')
        os.system('shutdown /r /t 30 /c "Reboot initialized by ESST"')
        return ''

    @staticmethod
    def show_cpu_usage_once():
        """Show CPU usage once"""
        LOGGER.debug('show cpu usage once')
        CTX.server_show_cpu_usage_once = True

    @staticmethod
    def show_cpu_usage_once_done():
        """Stop showing CPU usage once"""
        LOGGER.debug('show cpu usage once: done')
        CTX.server_show_cpu_usage_once = False

    @staticmethod
    def show_cpu_usage_start():
        """Starts showing CPU usage continuously"""
        LOGGER.debug('show cpu usage: start')
        CTX.server_show_cpu_usage = True

    @staticmethod
    def show_cpu_usage_stop():
        """Stops showing CPU usage continuously"""
        LOGGER.debug('show cpu usage: stop')
        CTX.server_show_cpu_usage = False

    @staticmethod
    def show_graph(days, hours, minutes):
        """Show resources usage graph"""
        def _callback(future):
            if future.result():
                DISCORD.send_file(future.result())
            else:
                LOGGER.warning('failed to create the graph')

        LOGGER.debug('show cpu usage: graph')
        make_history_graph(_callback, days, hours, minutes)



# coding=utf-8
"""
Inject more code into DCS 1.5.8 and 2.x (new login mechanism)
"""
import re

from esst import core, utils

LOGGER = core.MAIN_LOGGER.getChild(__name__)

# noinspection SpellCheckingInspection
INJECT_TEMPLATE = """function onShowMainInterface()
--print("--onShowMainInterface()---")
    if tooltipSkin_ == nil then
        tooltipSkin_ = Gui.GetTooltipSkin()
    else
        Gui.SetTooltipSkin(tooltipSkin_)
    end
    prepareMissionPath()
    mmw.setLastWallpaper()
    openReturnScreen()

    -- START DEDICATED CODE -  ADD FROM THIS LINE
    --
    if grgFirstRun == nil then
        grgFirstRun = true

        local net = require('net')
        local lfs = require('lfs')
        local Tools = require('tools')
        local mpConfig = Tools.safeDoFile(lfs.writedir() .. 'Config/serverSettings.lua', false)
        local dediConfig = Tools.safeDoFile(lfs.writedir() .. 'Config/dedicated.lua', false)

         if dediConfig and dediConfig.dedicated ~= nil and dediConfig.dedicated["enabled"] == true then
           net.set_name(dediConfig.dedicated["name"])
           net.start_server(mpConfig.cfg)
           net.log("Starting Dedicated Server...")
        end
    end
    --
    -- END DEDICATED CODE - ADD UP TO THIS LINE

end"""

RE_INJECT = re.compile(
    r"""^function onShowMainInterface\(\)\n\s*--.*(?:\n^.*?)*^end$""",
    re.MULTILINE
)


def inject_mission_editor_code() -> bool:
    """
    Injects code needed for the new login method in MissionEditor.lua

    Args:
        dcs_path: path to the DCS installation

    Returns:
        Bool indicating success of the operation

    """

    core.FS.ensure_path(core.FS.mission_editor_lua_file, 'mission editor lua file')

    LOGGER.debug(f'injecting MissionEditor.lua code in DCS installation: {core.FS.dcs_path}')
    if not core.FS.dcs_path.exists():
        raise FileNotFoundError(core.FS.dcs_path)

    LOGGER.debug(f'MissionEditor.lua path: {core.FS.mission_editor_lua_file}')
    if not core.FS.mission_editor_lua_file.exists():
        raise FileNotFoundError(core.FS.mission_editor_lua_file)

    LOGGER.debug('backing up MissionEditor.lua')
    utils.create_versioned_backup(core.FS.mission_editor_lua_file)

    LOGGER.debug('injecting code')
    output, count = RE_INJECT.subn(INJECT_TEMPLATE, core.FS.mission_editor_lua_file.read_text(encoding='utf8'))

    if count == 0:
        LOGGER.warning('no replacement made')
        return False

    LOGGER.debug('writing resulting file')
    core.FS.mission_editor_lua_file.write_text(output, encoding='utf8')
    return True



# coding=utf-8
"""
Manages config params for auto mission
"""

from elib.config import ConfigProp

NAMESPACE = 'AUTO_MISSION'


class AutoMissionConfig:
    """
    Manages config params for auto mission
    """
    auto_mission_github_token = ConfigProp(str, namespace=NAMESPACE, default='')
    auto_mission_github_owner = ConfigProp(str, namespace=NAMESPACE, default='')
    auto_mission_github_repo = ConfigProp(str, namespace=NAMESPACE, default='')



# coding=utf-8
"""
Manages config params for auto mission
"""

import inspect

from elib.config import BaseConfig, ConfigProp

from ._atis import ATISConfig
from ._auto_mission import AutoMissionConfig
from ._dcs import DCSConfig
from ._dcs_server import DCSServerConfig
from ._discord_bot import DiscordConfig
from ._remove_old_files import RemoveOldFile
from ._univers_radio import URConfig


# pylint: disable=too-many-ancestors
class ESSTConfig(BaseConfig,
                 DCSConfig,
                 DCSServerConfig,
                 AutoMissionConfig,
                 RemoveOldFile,
                 DiscordConfig,
                 URConfig,
                 ATISConfig):
    """
    Manages config params for auto mission
    """

    def get_context(self) -> dict:
        """

        Returns: dict context for Sentry

        """
        return {
            member: value
            for member, value in inspect.getmembers(self, lambda a: not inspect.ismethod(a))
            if not member.startswith('_')
        }

    def __init__(self):
        BaseConfig.__init__(self, 'esst')

    debug = ConfigProp(bool, default='false')
    restart = ConfigProp(str, default='')
    saved_games_dir = ConfigProp(str, default='')
    # noinspection SpellCheckingInspection
    sentry_dsn = ConfigProp(
        str,
        default='https://85518bcfd75a400eaf3821830ec1c4b2:a622d4e7a4ab4ec9ade873ad96b8d4aa@sentry.io/206995'
    )
    start_server_loop = ConfigProp(bool, default='true')
    start_dcs_loop = ConfigProp(bool, default='true')
    start_discord_loop = ConfigProp(bool, default='true')
    start_listener_loop = ConfigProp(bool, default='true')



# coding=utf-8
"""
Manages the server computer
"""

import datetime
import time

import psutil

from esst.commands import DISCORD
from esst.core import CTX, MAIN_LOGGER, ServerStatus
from esst.utils import now

LOGGER = MAIN_LOGGER.getChild(__name__)

# pylint: disable=too-few-public-methods


class App:
    """
    Manages the server computer
    """

    def __init__(self):

        if not CTX.start_server_loop:
            LOGGER.debug('skipping server loop init')
            return

        ServerStatus.logical_cpus = psutil.cpu_count()
        ServerStatus.physical_cpus = psutil.cpu_count(logical=False)
        ServerStatus.cpu_frequency = psutil.cpu_freq().max
        ServerStatus.total_memory = psutil.virtual_memory().total
        ServerStatus.swap_size = psutil.swap_memory().total
        ServerStatus.boot_time = datetime.datetime.fromtimestamp(
            psutil.boot_time()).strftime("%Y-%m-%d %H:%M:%S")

    @staticmethod
    def _update_status():
        while not CTX.exit:

            cpu_usage = psutil.cpu_percent(1)

            net_io = psutil.net_io_counters()
            bytes_sent = net_io.bytes_sent
            bytes_recv = net_io.bytes_recv

            ServerStatus.cpu_usage = cpu_usage
            if CTX.server_show_cpu_usage or CTX.server_show_cpu_usage_once:
                DISCORD.say(f'Server cpu usage: {cpu_usage}%')
                CTX.server_show_cpu_usage_once = False

            ServerStatus.used_memory = ServerStatus.total_memory - psutil.virtual_memory().free
            ServerStatus.mem_usage = round(
                ServerStatus.used_memory / ServerStatus.total_memory * 100, 2)
            ServerStatus.swap_used = psutil.swap_memory().used

            now_ = now()
            CTX.server_cpu_history.append((now_, cpu_usage))
            CTX.server_mem_history.append((now_, ServerStatus.mem_usage))

            # noinspection PyProtectedMember
            if ServerStatus.bytes_recv_ != 0:
                bytes_sent_ = bytes_sent - ServerStatus.bytes_sent_
                bytes_recv_ = bytes_recv - ServerStatus.bytes_recv_
                ServerStatus.bytes_sent = bytes_sent_
                ServerStatus.bytes_recv = bytes_recv_
                CTX.server_bytes_sent_history.append((now_, bytes_sent_))
                CTX.server_bytes_recv_history.append((now_, bytes_recv_))
            ServerStatus.bytes_recv_ = bytes_recv
            ServerStatus.bytes_sent_ = bytes_sent

            time.sleep(5)

    async def run(self):
        """
        Entry point of the loop
        """
        if not CTX.start_server_loop:
            LOGGER.debug('skipping server loop')
            return
        CTX.loop.run_in_executor(None, self._update_status)

        LOGGER.debug('end of Server computer loop')



# coding=utf-8
"""
Clean folders of old files
"""

from elib.config import ConfigProp


class RemoveOldFile:  # pylint: disable=too-few-public-methods
    """
    Clean folders of old files
    """

    remove_files = ConfigProp(list, default='')



# coding=utf-8
"""
Generates ATIS from METAR
"""

import threading
import typing
from pathlib import Path

import elib.tts
import emiz.weather

from esst import core

from ._atis_airfields import ALL_AIRFIELDS
from ._atis_identifier import get_random_identifier
from ._atis_objects import ATISForAirfield
from ._atis_status import Status
from ._univers_radio import URVoiceService, URVoiceServiceSettings

LOGGER = core.MAIN_LOGGER.getChild(__name__)


def _build_speech_for_airfield(airfield, wind_dir, speech_atis, ur_settings) -> ATISForAirfield:
    LOGGER.debug(f'{airfield.icao}: processing')
    atis_file = Path(f'{airfield.icao}.mp3').absolute()
    LOGGER.debug(f'{airfield.icao}: ATIS file path: {atis_file}')
    active_runway = airfield.get_active_runway(wind_dir)
    LOGGER.debug(f'{airfield.icao}: active runway: {active_runway.long_name()}')
    speech_intro = f'{airfield.icao}: ATIS for {airfield.name}'
    LOGGER.debug(f'{airfield.icao}: ATIS intro: {speech_intro}')
    speech_active_runway = f'Active runway {active_runway.long_name()}'
    LOGGER.debug(f'{airfield.icao}: active runway speech: {speech_active_runway}')
    information_identifier, information_letter = get_random_identifier()
    speech_information = f'Advise you have information, {information_identifier}, on first contact.'
    LOGGER.debug(f'{airfield.icao}: speech information: {speech_information}')
    full_speech = '. '.join([speech_intro, speech_atis, speech_active_runway, speech_information])
    LOGGER.debug(f'{airfield.icao}: full speech: {full_speech}')
    LOGGER.debug(f'{airfield.icao}: writing MP3 file for: {airfield.icao}')
    elib.tts.text_to_speech(full_speech, Path(atis_file), True)
    ur_settings.add_station(airfield)
    return ATISForAirfield(airfield.icao, active_runway, information_identifier, information_letter)


def generate_atis(metar_str: str,
                  include_icao: typing.List[str] = None,
                  exclude_icao: typing.List[str] = None):
    """
    Create MP3 from METAR
    """
    if not core.CFG.atis_create:
        LOGGER.info('skipping ATIS creation as per config')
        return
    URVoiceService.kill()
    LOGGER.info(f'creating ATIS from METAR: {metar_str}')
    LOGGER.debug('parsing METAR string')
    # noinspection SpellCheckingInspection
    metar_str = metar_str.replace('XXXX', 'UGTB')
    error, metar = emiz.weather.custom_metar.CustomMetar.get_metar(metar_str)
    if error:
        LOGGER.error('failed to parse METAR')
        raise RuntimeError(metar)
    wind_dir = metar.wind_dir.value()
    LOGGER.debug(f'wind direction: {wind_dir}')
    speech_atis = emiz.weather.AVWX.metar_to_speech(metar_str)
    core.CTX.atis_speech = speech_atis
    LOGGER.debug(f'ATIS speech: {speech_atis}')

    ur_settings = URVoiceServiceSettings()
    active_atis = {}

    if include_icao:
        include_icao = [icao.upper() for icao in include_icao]

    if exclude_icao:
        exclude_icao = [icao.upper() for icao in exclude_icao]

    threads = []
    for airfield in ALL_AIRFIELDS:
        if core.CTX.exit:
            break
        if include_icao and airfield.icao.upper() not in include_icao:
            LOGGER(f'{airfield.icao}: skipping (not included)')
            continue
        if exclude_icao and airfield.icao.upper() in exclude_icao:
            LOGGER(f'{airfield.icao}: skipping (excluded)')
            continue

        thread = threading.Thread(
            target=_build_speech_for_airfield,
            args=(airfield, wind_dir, speech_atis, ur_settings)
        )
        threads.append(thread)
        thread.start()
    for job in threads:
        job.join()

    list_of_active_icao = ', '.join(active_atis.keys())
    LOGGER.debug(f'generated {len(active_atis)} ATIS for: {list_of_active_icao})')
    Status.active_atis = active_atis

    LOGGER.debug('writing UR settings')
    ur_settings.write_settings_file()
    URVoiceService.start_service()



# coding=utf-8
"""
Globally available commands
"""

# noinspection PyUnresolvedReferences
from esst.discord_bot.commands import DISCORD
# noinspection PyUnresolvedReferences
from esst.listener.commands import LISTENER
# noinspection PyUnresolvedReferences
from esst.dcs.commands import DCS
# noinspection PyUnresolvedReferences
from esst.server.commands import SERVER



# coding=utf-8
"""
Retrieves info about currently running ATIS
"""

from ._atis_objects import ATISForAirfield as _ATISForAirfield
from ._atis_status import Status as _Status


def get_info_for_icao(icao: str) -> _ATISForAirfield:
    """Returns ATIS information for icao"""
    icao = icao.upper()
    return _Status.active_atis[icao]



# coding=utf-8
"""
Various helper functions
"""
import datetime
import os
import shutil
import typing
from pathlib import Path

import elib
import ipgetter
import pefile
import pkg_resources

from esst.core import MAIN_LOGGER, Status, FS
from .arg import arg
from .github import get_latest_release
from .remove_old_files import clean_all_folder

LOGGER = MAIN_LOGGER.getChild(__name__)


def external_ip():
    """
    Returns: external IP of this machine
    """
    return ipgetter.IPgetter().get_externalip()


def sanitize_path(path: typing.Union[str, Path]) -> str:
    """
    Sanitize a filesystem path

    Args:
        path: path to sanitize

    Returns: sanitized path as a string

    """
    return str(path).replace('\\', '/')


def _do_backup(original: Path, backup: Path):
    LOGGER.debug(f'checking for backup of {original.absolute()}')
    if not original.exists():
        LOGGER.debug(f'original does no exist, skipping backup: {original.absolute()}')
        return
    if not backup.exists():
        LOGGER.debug(f'creating backup of "{original.absolute()}" -> "{backup.absolute()}"')
        shutil.copy2(str(original.absolute()), str(backup.absolute()))
    else:
        LOGGER.debug(f'backup already exists: "{backup.absolute()}"')


def create_versioned_backup(file_path: typing.Union[str, Path], file_must_exist: bool = True):
    """
    Creates a backup of a file, with a "_backup_DCS-VERSION" suffix, if the backup does not exist yet

    Args:
        file_must_exist: fails if the file to be backed up does not exist
        file_path: file to backup

    """
    file_path = elib.path.ensure_file(file_path, must_exist=file_must_exist)
    backup_file = Path(file_path.parent, f'{file_path.name}_backup_{Status.dcs_version}')
    _do_backup(file_path, backup_file)


def create_simple_backup(file_path: typing.Union[str, Path], file_must_exist: bool = True):
    """
    Creates a backup of a file, with a "_backup_DCS-VERSION" suffix, if the backup does not exist yet

    Args:
        file_must_exist: fails if the file to be backed up does not exist
        file_path: file to backup

    """
    file_path = elib.path.ensure_file(file_path, must_exist=file_must_exist)
    backup_file = Path(file_path.parent, f'{file_path.name}_backup')
    _do_backup(file_path, backup_file)


def now():
    """

    Returns: epoch

    """
    return datetime.datetime.now().timestamp()


def read_template(template_name: str) -> str:
    """
    Reads a template file, getting it from the local install or from the package

    Args:
        template_name: name of the template file

    Returns: template file content

    """
    LOGGER.debug(f'reading template: {template_name}')
    template_path = os.path.join(os.path.dirname(__file__), 'templates', template_name)
    if not os.path.exists(template_path):
        LOGGER.debug('template not found, trying from pkg_resource')
        template_path = pkg_resources.resource_filename('esst', f'/dcs/templates/{template_name}')
    if not os.path.exists(template_path):
        raise FileNotFoundError(template_path)
    with open(template_path) as handle_:
        LOGGER.debug('returning template content')
        return handle_.read()


def get_esst_changelog_path() -> str:
    """

    Returns: changelog path

    """
    changelog_path = os.path.join(os.path.dirname(__file__), 'CHANGELOG.rst')
    if not os.path.exists(changelog_path):
        LOGGER.debug('changelog not found, trying from pkg_resource')
        changelog_path = pkg_resources.resource_filename(
            'esst', 'CHANGELOG.rst')
    if not os.path.exists(changelog_path):
        LOGGER.error('changelog not found')
        return ''

    return changelog_path


def get_dcs_log_file_path() -> str:
    """
    Returns: path to DCS log file
    """
    return os.path.join(FS.dcs_logs_dir, 'dcs.log')


class Win32FileInfo:
    """
    Gets information about a Win32 portable executable
    """

    def __init__(self, _path):

        self.__path = os.path.abspath(_path)
        self.__props = None
        self.__read_props()

    @property
    def comments(self):
        """Show Win32FileInfo field"""
        return self.__props.get('Comments')

    @property
    def internal_name(self):
        """Show Win32FileInfo field"""
        return self.__props.get('InternalName')

    @property
    def product_name(self):
        """Show Win32FileInfo field"""
        return self.__props.get('ProductName')

    @property
    def company_name(self):
        """Show Win32FileInfo field"""
        return self.__props.get('CompanyName')

    @property
    def copyright(self):
        """Show Win32FileInfo field"""
        return self.__props.get('LegalCopyright')

    @property
    def product_version(self):
        """Show Win32FileInfo field"""
        return self.__props.get('ProductVersion')

    @property
    def file_description(self):
        """Show Win32FileInfo field"""
        return self.__props.get('FileDescription')

    @property
    def trademark(self):
        """Show Win32FileInfo field"""
        return self.__props.get('LegalTrademarks')

    @property
    def private_build(self):
        """Show Win32FileInfo field"""
        return self.__props.get('PrivateBuild')

    @property
    def file_version(self):
        """Show Win32FileInfo field"""
        return self.__props.get('FileVersion')

    @property
    def fixed_version(self):
        """Show Win32FileInfo field"""
        return self.__props.get('fixed_version')

    @property
    def original_filename(self) -> str:
        """Show Win32FileInfo field"""
        return self.__props.get('OriginalFilename')

    @property
    def special_build(self) -> str:
        """Show Win32FileInfo field"""
        return self.__props.get('SpecialBuild')

    def __read_props(self):

        def _loword(dword):
            return dword & 0x0000ffff

        def _hiword(dword):
            return dword >> 16

        self.__props = {}

        try:
            pe_file = pefile.PE(self.__path)
        except pefile.PEFormatError as exc:
            raise ValueError(exc.value)
        else:
            # noinspection SpellCheckingInspection
            pvms = pe_file.VS_FIXEDFILEINFO.ProductVersionMS  # pylint: disable=no-member
            # noinspection SpellCheckingInspection
            pvls = pe_file.VS_FIXEDFILEINFO.ProductVersionLS  # pylint: disable=no-member
            self.__props['fixed_version'] = '.'.join(
                map(str, (_hiword(pvms), _loword(pvms), _hiword(pvls), _loword(pvls)))
            )
            for file_info in pe_file.FileInfo:
                if file_info.Key == b'StringFileInfo':
                    for str_table in file_info.StringTable:
                        for entry in str_table.entries.items():
                            self.__props[entry[0].decode(
                                'latin_1')] = entry[1].decode('latin_1')
